#!/usr/bin/env python3
"""
LTE Daemon with autoswitch feature - designed for Unipi PLCs
    Created on 3 Jun 2019 - version 0.26
@author: martyy <triska@unipi.technology>
"""

import argparse
import configparser
import datetime
import hashlib
import ipaddress
import logging
import logging.handlers
import os
import random
import re
import signal
import socket
import stat
import string
import struct
import subprocess
import sys
import time
import traceback
from collections.abc import Callable
from enum import Enum
from time import sleep

import serial

# Network states
WAN_STATE_INIT = 0
WAN_STATE_LTE_ON = 1
WAN_STATE_LTE_OFF = 2
WAN_STATE_ETH_ON = 3
WAN_STATE_LTE_ON_ETH_OFF = 4
WAN_STATE_LTE_OFF_ETH_ON = 5
WAN_STATE_LTE_ON_ETH_ON = 6
WAN_STATE_LTE_OFF_ETH_OFF = 7
WAN_STATE_ETH_OFF = 8
WAN_STATE_MONITOR_OFF = 9

# Specific error codes (no all known errors listed)
ERROR_CODE_UNSPEC = 0
ERROR_CODE_NO_SIM = 1
ERROR_CODE_SIM_ERROR = 2
ERROR_CODE_WRONG_PIN = 3

# Actions usually performed in non-standard cases
ACTION_NOTHING = 0
ACTION_CONTINUE = 1
ACTION_REINIT = 2
ACTION_SOFTRESTART = 3
ACTION_HARDRESTART = 4
ACTION_SUSPEND = 5
ACTION_HALT = 6

# Default values, can be overridden in the .conf file
DEFAULT_LOGGING_LEVEL = "INFO"
DEFAULT_LOG_PATH = None  # Log path is none -> use syslog instead of an user log file
DEFAULT_LOGFILE_SIZE = 131072  # 128 kiB
RESCUE_THRESHOLD = 350  # Not-success ping to modem restart

# Global definitions
DAEMON_VERSION = "0.60.2"  # _CIREPL_VERSION_
DEFAULT_MODEM_PORT = "/run/unipi-plc/by-sys/lte.1/tty_mng"  # Symlink to ttyACM0
DEFAULT_NIC_FILE = "/run/unipi-plc/by-sys/lte.1/interface"
DEFAULT_ENABLE_RFKILL = "/run/unipi-plc/by-sys/lte.1/rfkill"

CONFILE_PATH = "/etc/unipi/lte.conf"
PIDFILE_PATH = "/run/unipilte.pid"
NIC_CONF_PATH_INTERFACES = "/etc/network/interfaces.d/"
NIC_CONF_PATH_NETWORKD = "/etc/systemd/network/"

DEFAULT_LED_PATHS = {
    'sim': "/run/unipi-plc/by-sys/lte.1/led.sim",
    'net': "/run/unipi-plc/by-sys/lte.1/led.net",
    'sig': "/run/unipi-plc/by-sys/lte.1/led.sig",
    'int': "/run/unipi-plc/by-sys/lte.1/led.int"
}

PLATFORM_SPEC = {
    "ID-LTE-1": {
        'rfkill_enable': "/run/unipi-plc/by-sys/lte.1/rfkill",
        'leds': {
            'sim': "!/run/unipi-plc/by-sys/lte.1/led.int.r",
            'net': "!/run/unipi-plc/by-sys/lte.1/led.sig.r",
            'sig': "/run/unipi-plc/by-sys/lte.1/led.sig.g",
            'int': "/run/unipi-plc/by-sys/lte.1/led.int.g",
     }},
}

SUPPORTED_USB_MODEMS = {
    (0x2c7c, 0x0901): "EC200U",
    (0x2c7c, 0x6005): "EC200A",
    (0x2c7c, 0x0901): "EG915U",
    (0x2c7c, 0x6002): "EG912Y",
}

# Super-global variables
current_config: dict = {}


# Inherited exceptions
class ModemCreateInstanceException(Exception):
    pass


class TimeoutException(Exception):
    pass


class ModemConnectionError(Exception):
    pass


class ModemErrorException(Exception):
    pass


class LEDErrorException(Exception):
    pass


class NewSettings(Exception):
    pass


class LogMaster:
    LOG_FORMAT = '%(asctime)s - %(levelname)s - %(message)s'
    LOG_ID = 'unipi-lte'

    @classmethod
    def getLogger(cls):
        return logging.getLogger() #cls.LOG_ID)

    @classmethod
    def logger_init(cls, console:bool) -> None:
        """
        Main logger initializer.
        During the start phase, all logs goes to syslog (journal)
        If a logging to file is configured in the config file, it will be switched later.
        """
        logger = logging.getLogger() #cls.LOG_ID)
        if not console:
            syslog_handler = logging.handlers.SysLogHandler(address='/dev/log')
            syslog_handler.ident = f"{cls.LOG_ID}: "
        else:
            syslog_handler = logging.StreamHandler(sys.stdout)
            formatter = logging.Formatter(cls.LOG_FORMAT)
            syslog_handler.setFormatter(formatter)

        logger.addHandler(syslog_handler)
        logger.setLevel(logging.INFO)

    @classmethod
    def logger_swich_to_file(cls, _log_path: str, _log_size: int) -> None:
        """ Switch from syslog/systemd journal logging to a rotating file logger. """

        logger = logging.getLogger() #cls.LOG_ID)
        # Remove all existing handlers (e.g., syslog or systemd journal)
        for handler in logger.handlers[:]:
            logger.removeHandler(handler)
            handler.close()

        # Create a new file-based handler
        filelog_handler = logging.handlers.RotatingFileHandler(filename=_log_path, maxBytes=_log_size, backupCount=1)

        log_formatter = logging.Formatter(cls.LOG_FORMAT)
        filelog_handler.setFormatter(log_formatter)

        logger.addHandler(filelog_handler)


class PingMan:
    ip_cache: dict[str, str] = {}

    def __init__(self, ping_timeout: int, ping_target: str):

        self.do_ping: Callable[[str], bool]

        try:
            ipaddress.IPv4Address(ping_target)
            self.do_ping = lambda interface: self.ping_ip4(ip_address=ping_target, interface=interface,
                                                           timeout=ping_timeout)
            logging.info('Ping to IPv4 address: ' + ping_target)
            return
        except ValueError:
            pass

        try:
            ipaddress.IPv6Address(ping_target)
            #print("IPV6")
            self.do_ping = lambda interface: self.ping_ip6(ping_target)
        except ValueError:
            pass

        if '.' in ping_target:
            self.do_ping = lambda interface: self.ping_domain(ping_target=ping_target, interface=interface,
                                                              timeout=ping_timeout)

            logging.info('Ping to domain name: ' + ping_target)
        else:
            logging.error("Invalid ping target, use domain name or IP address")
            raise ValueError

    @staticmethod
    def resolve_domain(domain):
        """Resolve the domain name to an IP address."""
        try:
            ip_address = socket.gethostbyname(domain)
            logging.info(f"[INFO] Domain '{domain}' resolved to: {ip_address}")
            return ip_address
        except socket.gaierror as error:
            logging.error(f"[ERROR] Failed to resolve domain '{domain}': {error}")
            return None

    @classmethod
    def ping_domain(cls, ping_target, interface=0, timeout=0) -> bool:

        if ping_target not in cls.ip_cache:
            ip_add = cls.resolve_domain(ping_target)
            if ip_add is None:
                return False
            else:
                cls.ip_cache[ping_target] = ip_add

        if (ret := cls.ping_ip4(cls.ip_cache[ping_target], interface, timeout)) is False:
            cls.ip_cache.pop(ping_target)

        return ret

    @staticmethod
    def ping_ip4(ip_address, interface=0, timeout=0) -> bool:
        """Ping the given IP address a specified number of times."""

        command = ['ping', '-c', '1',  # Send 1 ping packet
            '-W', str(timeout),  # Set timeout (in seconds) for the ping response
            '-I', str(interface),  # Specify the interface to use
            '-s', '32',  # Specify the packet size (optional)
            ip_address]

        try:
            # Run the ping command and capture the output.
            logging.debug(" ".join(command))
            result = subprocess.run(command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            return result.returncode == 0
        except subprocess.CalledProcessError:
            return False

    @staticmethod
    def ping_ip6(ip_address) -> bool:
        raise NotImplementedError("Method ping_ip6 not implemented")


class LedServer(object):
    led_states = {"off": "0", "on": "1", "off_inverted": "1", "on_inverted": "0", "blink_fast": "150",
                  "blink_slow": "2000"}

    def __init__(self, led_paths: dict[str: str]):  # Check presence of all leds

        self.led_inverted = {}
        self.leds = {}
        self.led_cache: {str: str} = {}

        logging.debug("LedServer init: " + str(led_paths))

        for led, path in led_paths.items():

            if "!" in path:
                self.led_inverted[led] = True
                self.leds[led] = path.strip('!')
            else:
                self.led_inverted[led] = False
                self.leds[led] = path

            if not os.path.exists(self.leds[led]):
                raise LEDErrorException(f"Led path not valid {self.leds[led]}")

        for led in self.leds:  # Init da cache
            self.led_cache[led] = ""
            self.led_cache[led + '_duty'] = 0

        self.all_off()

    @staticmethod
    def __write_led_value(path: str, value: str):
        with open(path, "w") as _file:
            _file.write(str(value))

    def set(self, led, state, bypass_inversion = False):  # Set state of LTE coupled LEDs

        try:
            match state:
                case _ if self.led_cache.get(led) == state:
                    # Return early if the led state has not changed
                    return
                case "on" | "off":
                    # Set trigger to "none" for on/off states
                    LedServer.__write_led_value(f"{self.leds[led]}/trigger", "none")

                    # Adjust state if led is inverted
                    if self.led_inverted.get(led, False) is True and bypass_inversion is False:
                        state = state + "_inverted"

                    # Write brightness value
                    LedServer.__write_led_value(f"{self.leds[led]}/brightness", self.led_states[state])

                case "blink_fast" | "blink_slow":
                    # Set trigger to "timer" for blink states
                    LedServer.__write_led_value(f"{self.leds[led]}/trigger", "timer")

                    # Write delay values for blinking
                    delay_value = self.led_states[state]
                    LedServer.__write_led_value(f"{self.leds[led]}/delay_on", delay_value)
                    LedServer.__write_led_value(f"{self.leds[led]}/delay_off", delay_value)
                case _ as _E:
                    raise LEDErrorException(f"Led state not valid: {_E}")

        except KeyError:
            raise LEDErrorException(f"LED {led} is not defined.")
        except Exception as _E:
            raise LEDErrorException(f"Cannot set LED {led} to state {state}: {_E}")

        self.led_cache[led] = state

    def duty(self, led: str, duty: int):  # Set LED duty cycle (used as a signal quality indicator)
        if self.led_cache[led + '_duty'] == duty and self.led_cache[led] == "duty":
            return

        per = (3000 / 31) * (31 - duty)
        t_on = int(per)  # 0-930
        t_off = int(per)

        LedServer.__write_led_value(f"{self.leds[led]}/trigger", "timer")
        LedServer.__write_led_value(f"{self.leds[led]}/delay_on", str(t_on))
        LedServer.__write_led_value(f"{self.leds[led]}/delay_off", str(t_off))

        self.led_cache[led + '_duty'] = duty
        self.led_cache[led] = "duty"

    def all_off(self):  # Switch of all LEDs
        for led in self.leds:
            self.set(led, 'off', bypass_inversion=True)

    def all_on(self):  # Switch of all LEDs
        for led in self.leds:
            self.set(led, 'on')

    def all_blink(self, fast=True):
        mode = 'off'
        for led in self.leds:
            self.set(led, mode)
        mode = 'blink_fast' if fast else 'blink_slow'
        for led in self.leds:
            self.set(led, mode)


class LTEModemBase(object):
    AT_OK_REPLY = b"OK"
    # Keep the regexs pre-compiled
    REGEX_CSQ = re.compile(r"\+CSQ:\s*(\d+),(\d+)")
    REGEX_COP = re.compile(r"\+COPS:\s*\d+,\d?,\"(.*)\",(\d?)")

    def __init__(self, at_port_path: str = DEFAULT_MODEM_PORT, nic_name_file=DEFAULT_NIC_FILE):
        logging.debug("Creating LTEModemBase")

        self.serial = None
        self.at_tty = at_port_path
        self.nic_name = nic_name_file

    def iface_init(self):
        if not stat.S_ISCHR(os.stat(self.at_tty).st_mode):  # Check if modem virtual serial interface exists
            raise ModemErrorException("Modem serial port {} is not a valid character device".format(self.at_tty))
        if hasattr(self.serial, 'open'):
            self.serial.close()

        try:
            self.serial = serial.Serial(port=self.at_tty, baudrate=115200, timeout=0.3, parity=serial.PARITY_NONE,
                                        rtscts=False)
        except Exception as _E:
            raise ModemErrorException("Cannot open virtual serial port {}".format(self.at_tty))

        self.serial.flush()
        self.serial.reset_output_buffer()
        self.serial.reset_input_buffer()

        current_config['lte_iface'] = get_lte_iface_name(self.nic_name, create_config_file=True)

        if not current_config['lte_iface']:
            raise ModemErrorException("No network interface found in file {}".format(self.nic_name))

    def iface_wait_for(self, disconnect_first=False):

        if disconnect_first is True:
            while True:
                logging.debug("Waiting for modem disconnect...")
                try:
                    stat.S_ISCHR(os.stat(self.at_tty).st_mode)
                except:
                    break
                time.sleep(0.3)

            logging.debug("Disconnected")

        while True:
            logging.debug("Waiting for modem reconnect...")
            try:
                stat.S_ISCHR(os.stat(self.at_tty).st_mode)
            except:
                time.sleep(0.3)
                continue
            break

        logging.debug("Reconnected")

    def iface_flush(self):
        self.serial.read(4065)  # Dummy read to flush the input buffer

    def query(self, query, bytes_expected=569):
        retry_cnt = 3

        while retry_cnt > 0:
            try:
                self.serial.flushOutput()  # Flush da buffer
                self.serial.flushInput()  # Flush da buffer
                # time.sleep(0.1)
                self.serial.write(query.encode())  # Write op. is blocking - may take longer !
                chars = self.serial.read(bytes_expected)  # Read with previously defined timeout
            except Exception as _e:  # May occur when other app uses modem's serial line
                raise ModemErrorException("Modem I/O error " + str(_e))
            # logging.debug("RECEIVED {} bytes: {}".format(len(chars),chars))
            if b"ERROR" in chars:
                raise ModemErrorException(chars)
            if len(chars) > 0:  # No reply received
                return chars

            retry_cnt = retry_cnt - 1
            if len(query) > 4:  # Supress AT\r\n polling during startup sequence
                logging.info("More attempts for " + str(query).strip('\r\n'))

        raise TimeoutException

    def query_extended_timeout(self, query, timeout, bytes_expected=569):

        timeout_bckp = self.serial.timeout
        self.serial.timeout = timeout
        try:
            ret = self.query(query, bytes_expected)
        finally:
            self.serial.timeout = timeout_bckp
        return ret

    def init(self):
        pass

    def get_sig_metrics(self) -> tuple:
        pass

    def get_operator_nettype(self) -> tuple:
        pass

    def get_ip_addresses(self) -> tuple[str, str]:
        pass

    def sw_restart(self):
        pass

    def resolve_error(self, exc, leds) -> tuple[int, int]:
        pass


class LTEModemPWRControl(object):

    def __init__(self, enable_rfkill_path: str = DEFAULT_ENABLE_RFKILL):

        self.enable_path = enable_rfkill_path

        if not os.path.isdir(self.enable_path):
            raise ModemCreateInstanceException(
                f"Rfkill power control not found ({enable_rfkill_path})")

    def enable(self):
        with open(self.enable_path + "/soft", "w") as f:
            f.write("0")

    def disable(self):
        with open(self.enable_path + "/soft", "w") as f:
            f.write("1")

    def hw_restart(self):
        with open(self.enable_path + "/soft", "w") as f:
            f.write("1")
        sleep(3)
        with open(self.enable_path + "/soft", "w") as f:
            f.write("0")
        sleep(3)


class LTEModemPlatform(object):

    def __init__(self, modem: LTEModemBase | None, leds: LedServer, power_ctrl: LTEModemPWRControl):

        self.modem: LTEModemBase = modem
        self.leds: LedServer = leds
        self.pwrctrl: LTEModemPWRControl = power_ctrl

    def do_cleanup(self):

        logging.info("Modem platform ending...")
        if self.modem is not None and self.modem.serial:
            logging.info("Closing serial line...")
            self.modem.serial.close()

        if self.leds:
            self.leds.all_off()

    def init(self):
        self.init_interface()  # Part of generic modem
        self.modem.init()  # Part of specific modem
        self.leds.set('sim', 'on')

    def init_interface(self, retry_counter=5):

        exc = None
        delay = 8

        retry_cnt = retry_counter / 2

        while retry_cnt > 0:
            try:
                self.modem.iface_init()
                self.leds.all_off()
                logging.debug("Init interface OK")
                return

            except Exception as _E:
                logging.debug(str(_E) + " - attempt {}/{}".format(retry_cnt, retry_counter))
            retry_cnt -= 1
            sleep(1)

        self.pwrctrl.enable()
        sleep(delay)

        while retry_counter > 0:
            try:
                self.modem.iface_init()
                self.leds.all_off()
                logging.debug("Init interface OK")
                return
            except Exception as _E:
                logging.debug(str(_E))

            retry_counter -= 1
            if retry_counter < 2:
                logging.info("Modem init retry with HW restart...{} attempts left".format(retry_counter))
                # self.restart(soft = False)
                self.pwrctrl.hw_restart()
                delay = 10  # Extend delay to wait for modem boot
            else:
                logging.info("Modem init retry...{} attempts left".format(retry_counter))
                self.pwrctrl.enable()

            sleep(random.random() * 10 + delay)
            exc = _E

        self.leds.all_blink(fast=False)

        raise ModemCreateInstanceException(str(exc))

    def update_led(self, is_registered, rssi, wan_state):

        if is_registered is True:  # Registered in the network
            self.leds.set('net', 'on')
        else:
            self.leds.set('net', 'off')

        if rssi == 99:
            self.leds.set('sig', 'off')
        else:
            self.leds.duty('sig', rssi)

        if current_config['ping']['autoswitch']:
            if wan_state == WAN_STATE_LTE_ON_ETH_ON:  # Connection to the internet (or other WAN network)
                self.leds.set('int', 'on')
            elif wan_state in {WAN_STATE_LTE_ON_ETH_OFF, WAN_STATE_LTE_ON}:
                self.leds.set('int', 'blink_fast')
            elif wan_state in {WAN_STATE_LTE_OFF_ETH_ON, WAN_STATE_ETH_ON}:
                self.leds.set('int', 'blink_slow')
            else:  # OFFLINE
                self.leds.set('int', 'off')
        else:  # No autoswitch
            if wan_state in {WAN_STATE_ETH_ON, WAN_STATE_LTE_ON}:
                self.leds.set('int', 'on')
            else:
                self.leds.set('int', 'off')

    def go_halt_suspend(self, errcode=ERROR_CODE_UNSPEC):

        if errcode in (ERROR_CODE_NO_SIM, ERROR_CODE_SIM_ERROR):
            self.leds.all_off()
            self.leds.set('sim', 'blink_slow')
        elif errcode == ERROR_CODE_WRONG_PIN:
            self.leds.all_off()
            self.leds.set('sim', 'blink_fast')
        else:
            self.leds.all_blink(fast=False)

    def get_sig_metrics(self):
        return self.modem.get_sig_metrics()

    def get_operator_nettype(self):
        return self.modem.get_operator_nettype()

    def get_ip_addresses(self):
        return self.modem.get_ip_addresses()

    def resolve_error(self, exc):
        return self.modem.resolve_error(exc, self.leds)

    def restart_hw(self):
        return self.pwrctrl.hw_restart()

    def restart(self, soft=True):
        if soft is False:
            return self.pwrctrl.hw_restart()
        else:
            self.modem.sw_restart()

        self.modem.iface_wait_for(disconnect_first=True)


class USBModemDevice:
    def __init__(self, device_name: str, _vid: int, _pid: int) -> None:
        """
        Initializes a USB device object with the given path, VID, and PID.

        :param device_name: The device path in the system (e.g., /sys/bus/usb/devices/1-1)
        :param _vid: Vendor ID of the USB device
        :param _pid Product ID of the USB device
        """
        super().__init__()

        self._pid = 0
        self._vid = 0

        self.device_name: str = device_name
        self.vid = _vid
        self.pid = _pid

    def __str__(self) -> str:
        """
        Returns a string representation of the USB device showing its VID and PID.
        """
        return f"USB modem {self.device_name} ({self.vid}:{self.pid})"

    @property
    def vid(self) -> str:
        return f"{self._vid:04x}"

    @property
    def pid(self) -> str:
        return f"{self._pid:04x}"

    @vid.setter
    def vid(self, value: str | int):
        if isinstance(value, int):
            self._vid = value
        elif isinstance(value, str):
            self._vid = int(value, 16)

    @pid.setter
    def pid(self, value: str | int):
        if isinstance(value, int):
            self._pid = value
        elif isinstance(value, str):
            self._pid = int(value, 16)


class LTEModemQuectelGeneric(LTEModemBase):
    REGEX_WAN_IP = re.compile(r"\+CGPADDR:\s*\d+,\"([^\"]*)\"")
    REGEX_LOCAL_IP = re.compile(r'(?<=inet )((?:\d+\.){3}\d+)')

    rept_rst_cnt = 0
    last_rst_time = time.time()

    def __init__(self, at_port_path: str = DEFAULT_MODEM_PORT, nic_name_file: str = DEFAULT_NIC_FILE,
                 usb_device: USBModemDevice = None):
        super().__init__(at_port_path=at_port_path, nic_name_file=nic_name_file)
        self.usb_device = usb_device

    def _resolve_apn(self):
        current_state = self.query("AT+CGDCONT?\r\n")

        # +CGDCONT: <cid>,<PDP_type>,<APN>,<PDP_addr>,<data_comp>,<head_comp>,<IPv4_addr_alloc>,<request_type>
        at_content = "1,\"{}\",\"{}\",\"\",0,0,1,0".format(current_config['conn_type'],
                                                   current_config['apn'])
        subcontent = ','.join(at_content.split(',')[:3])

        if subcontent.encode() in current_state:
            logging.debug("APN remains the same : " + current_config['apn'])
        else:
            logging.info("Changing APN to " + at_content)

            # Activate Airplane mode - alternative way to enable changing APN
            # self.query("AT+CFUN=4\r\n".format(at_content)) # use Airplane mode
            # self.query("AT+CFUN=1\r\n".format(at_content)) # use Standard mode

            # Disconnect from network
            self.query("AT+QNETDEVCTL=0\r\n".format(at_content))

            # Deactivate context
            self.query("AT+CGACT=0,1\r\n".format(at_content))

            # Set context
            self.query("AT+CGDCONT={}\r\n".format(at_content))

            # Activate context again
            self.query("AT+CGACT=1,1\r\n".format(at_content))

            raise NewSettings  # Restart is needed

    def init(self):
        super().init()

        MAX_RETRIES = 5
        RETRY_INTERVAL = 2  # Time to wait between retries when invalid reply
        RETRY_INTERVAL_ON_TIMEOUT = 0.5  # Time to wait when timeout occurs

        self.iface_flush()

        retry_counter = MAX_RETRIES

        while retry_counter > 0:
            try:
                response = self.query("AT\r\n", 10)  # Check if modem responds to AT commands

                if self.AT_OK_REPLY in response:
                    break  # Exit loop on successful reply
                else:
                    logging.debug(f"Invalid reply from modem, retries left {retry_counter}...")
                    sleep(RETRY_INTERVAL)

            except TimeoutException:
                logging.debug(f"No reply from modem, retries left {retry_counter}...")
                sleep(RETRY_INTERVAL_ON_TIMEOUT)

            retry_counter -= 1  # Decrement retry count after each attempt

        if retry_counter == 0:
            raise TimeoutException("Modem did not respond after multiple attempts.")

        while self.AT_OK_REPLY not in self.query("AT+CMEE=1\r\n",
                                                 17):  # Set extended error report (see CME codes in manual)
            sleep(1)

        while True:  # Manage pin stuff
            try:
                chars = self.query("AT+CPIN?\r\n")
                if (b"SIM PIN" in chars):
                    if (current_config['pin'] != None):
                        self.query("AT+CPIN={}\r\n".format(current_config['pin']))
                    else:  # Prevent to inserting invalid PIN if undefined in config file
                        logging.debug("No PIN code defined in {}".format(CONFILE_PATH))
                        raise ModemErrorException(
                            "ERROR: 16")  # This is the same ERR code as generated by the LTE modem.

                    continue
                elif (b"SIM PUK" in chars):
                    logging.warning("PUK code needed! Use AT Command interface to unlock the SIM.")
                    raise ModemErrorException("ERROR: 12")  # This is the same ERR code as generated by the LTE modem.
                elif (b"READY" in chars):
                    logging.info("Pin resolved")
                    break
                else:
                    logging.debug("UNEXPECTED DATA RECEIVED len({}) {}".format(len(chars), chars))
            except Exception as exc:
                if "ERROR: 14" in str(exc):
                    logging.info("Sim busy, waiting...")
                    continue
                else:
                    raise exc

        self._resolve_apn()

        current_state = self.query("AT+QNETDEVCTL?\r\n")

        logging.debug("Current connection status QNETDEVCTL : " + current_state.decode())
        if ("0,0".encode() in current_state):
            logging.debug("Enabling connection")
            current_state = self.query("AT+QNETDEVCTL=3,1,0\r\n")  # Parameter 3 = connection auto keep-on
        else:
            logging.debug("Already connected")

        ConnectionGuard.try_interface_up(current_config['lte_iface'])
        # os.system("ip link set {} up".format(current_config['lte_iface']))
        logging.info("Modem init complete")

    def get_sig_metrics(self):
        ret = self.query("AT+CSQ\r\n")
        res = self.REGEX_CSQ.search(str(ret))
        if res is None or len(res.groups()) != 2:
            logging.info("Cannot parse signal quality, received: {}".format(ret))
            return (99, 0)
        return int(res.group(1)), int(res.group(2))

    def get_operator_nettype(self):
        ret = self.query("AT+COPS?\r\n")
        res = LTEModemBase.REGEX_COP.search(str(ret))
        if res is None or len(res.groups()) != 2:
            # logging.info("Cannot parse operator info, received: {}".format(ret))
            return None, None

        return res.group(1), res.group(2)

    def get_ip_addresses(self) -> tuple[str, str]:

        _ip = self.REGEX_WAN_IP.search(str(self.query_extended_timeout("AT+CGPADDR=1\r\n", timeout=3.0)))
        if isinstance(_ip, re.Match):
            _ip = _ip.group(1)

        ret = subprocess.check_output("ip addr show " + current_config['lte_iface'], shell=True, text=True)
        _priv_ip = self.REGEX_LOCAL_IP.search(str(ret))
        if isinstance(_priv_ip, re.Match):
            _priv_ip = _priv_ip.group(1)

        return _ip, _priv_ip

    def resolve_error(self, exc, leds):

        _action = ACTION_NOTHING
        _errcode = ERROR_CODE_UNSPEC
        errstring = str(exc)

        if isinstance(exc, TimeoutException):
            logging.error("Modem reply timeout")
            _action = ACTION_HARDRESTART
        elif "ERROR: 10" in errstring:
            logging.error("SIM Not inserted")
            leds.set('sim', 'blink_slow')
            _errcode = ERROR_CODE_NO_SIM
            _action = ACTION_SOFTRESTART
        elif "ERROR: 13" in errstring:
            logging.error("SIM Failure")
            leds.set('sim', 'blink_slow')
            _errcode = ERROR_CODE_SIM_ERROR
            _action = ACTION_SOFTRESTART
        elif "ERROR: 14" in errstring:
            logging.info("SIM Busy")
            leds.set('sim', 'blink_slow')
            _errcode = ERROR_CODE_SIM_ERROR
            _action = ACTION_NOTHING
        elif "ERROR: 16" in errstring:
            logging.error("Incorrect or undefined password (PIN). Going to sleep...")
            leds.set('sim', 'blink_fast')
            _errcode = ERROR_CODE_WRONG_PIN
            _action = ACTION_HALT
        elif "ERROR: 12" in errstring:
            logging.error("SIM Locked, PUK needed. Going to sleep...")
            leds.set('sim', 'blink_fast')
            _errcode = ERROR_CODE_WRONG_PIN
            _action = ACTION_HALT
        elif "Modem I/O error" in errstring:
            logging.error("{}, performing HARD restart".format(errstring))
            _action = ACTION_HARDRESTART
        elif "ERROR: 569" in errstring:
            logging.error("An error with code 569")
            _action = ACTION_NOTHING
        else:
            logging.error("Unhandled error occured: " + errstring)
            _action = ACTION_SOFTRESTART

        # Time filter - if more than THRESHOLD restarts occurs with minimal delay, daemon goes to SUSPEND
        if _action == ACTION_HARDRESTART or _action == ACTION_SOFTRESTART:
            if (time.time() - LTEModemQuectelGeneric.last_rst_time) < 60.0:
                logging.info(
                    "Too rapid restart request, {} left to SUSPEND ".format(10 - LTEModemQuectelGeneric.rept_rst_cnt))
                LTEModemQuectelGeneric.rept_rst_cnt = LTEModemQuectelGeneric.rept_rst_cnt + 1

            if LTEModemQuectelGeneric.rept_rst_cnt > 10:
                logging.warning("Maximum number of resets reached, going to SUSPEND".format())
                LTEModemQuectelGeneric.rept_rst_cnt = 0
                LTEModemQuectelGeneric.last_rst_time = 0
                _action = ACTION_SUSPEND

            LTEModemQuectelGeneric.last_rst_time = time.time()

        return _action, _errcode

    def sw_restart(self):

        chars = self.query_extended_timeout("AT+CFUN=1,1\r\n", 3.0, 15)
        self.serial.close()


class LTEModemQuectelOLD(LTEModemQuectelGeneric):
    ''' This setting was used for EC912Y and EC200A
        but they works with new common setting too
    '''
    def _resolve_apn(self):

        current_state = self.query("AT+CGDCONT?\r\n")

        at_content = "1,{},{},\"0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0\",0,0,0,0,1,1".format(current_config['conn_type'],
                                                                                      current_config['apn'])
        if at_content.encode() in current_state:
            logging.debug("APN remains the same : " + current_config['apn'])
        else:
            logging.info("Changing APN to " + at_content)

            self.query("AT+CGDCONT={}\r\n".format(at_content))

            raise NewSettings  # Restart is needed


"""
Oldies - cleanup needed TODO
"""


def sig_handler(signum, frame):
    logging.info("Exiting after received signal {}".format(signum))
    Stats.set_default_values()
    sys.exit(0)


# Graceful exit with log event generated
# Parse of config file with int/bool data types conversion and default values


# Some of the network providers can assign WAN IP from the local range(s).
# In the case of IP conflict on both subnets, changes the appropriate IPs.


def resolve_subnets(ip_wan, ip_private, subnet):
    v1_ip = "192.168.101.1"  # The first variant
    v1_dhcp = "192.168.101.100"  # The first variant DHCP range
    v2_ip = "10.11.12.1"  # Second variant
    v2_dhcp = "10.11.12.100"  # Second variant DHCP range
    new_subnet = "255.255.255.0"

    def ip_to_binary(ip):  # Helper function converting string representation of given ip to binary integer
        octet_list_int = ip.split(".")
        octet_list_bin = [format(int(i), '08b') for i in octet_list_int]
        binary = ("").join(octet_list_bin)
        return binary

    subnet_wan = int(ip_to_binary(ip_wan), 2) & int(ip_to_binary(subnet), 2)  # Bitwise AND of IP and SUBNET
    subnet_private = int(ip_to_binary(ip_private), 2) & int(ip_to_binary(subnet), 2)  # Bitwise AND of IP and SUBNET

    logging.debug("WAN IP {}, Private IP {}; {} ".format(ip_wan, ip_private,
                                                        "CONFLICT" if subnet_private == subnet_wan else "NON-CONFLICT"))

    if subnet_wan == subnet_private:  # According to the doc from UBlox, ip range of LAN and WAN must be different
        logging.info("Resolving IP conflict")
        if ip_private == v1_ip:
            return {'ip': v2_ip, 'dhcp': v2_dhcp, 'mask': new_subnet}
        else:
            return {'ip': v1_ip, 'dhcp': v1_dhcp, 'mask': new_subnet}

    return None


# Get path of the interface config file used
def get_lte_interface_confile_path(iface):
    global NIC_CONF_PATH_INTERFACES
    global NIC_CONF_PATH_NETWORKD

    if os.path.isfile(NIC_CONF_PATH_INTERFACES + "interfaces"):  # TODO Use method
        logging.debug("Debian interfaces is used as network manager.".format(iface))
        return NIC_CONF_PATH_INTERFACES + iface
    else:
        logging.debug("Systemd-networkd is used as network manager.".format(iface))  # TODO - check netwd running
        return NIC_CONF_PATH_NETWORKD + "11-unipi-lte.network"


def check_iface_config_file(iface):
    current_nic_conf_path = get_lte_interface_confile_path(iface)

    if "systemd" in current_nic_conf_path:
        IFACE_FILE_CONTENT = f"""
[Match]
Name={iface}

[Network]
DHCP=yes

[DHCPv4]
UseRoutes=false

[Route]
Destination=0.0.0.0/0
Gateway=192.168.43.1
Metric=9999
GatewayOnLink=true
"""
    else:
        IFACE_FILE_CONTENT = "allow-hotplug {0}\niface {0} inet dhcp\n".format(iface)

    if not os.path.isfile(current_nic_conf_path) or IFACE_FILE_CONTENT not in open(current_nic_conf_path).read():

        file = open(current_nic_conf_path, 'w+')
        file.write(IFACE_FILE_CONTENT)
        file.close()
        logging.info("Created/corrected {} iface file".format(iface))
        subprocess.call("/bin/networkctl reload", shell=True)
    else:
        logging.debug("Interface {} iface file is OK and up-to date.".format(iface))


# Checks if config file (depends on network manager used) exists. If no, creates it.


def get_lte_iface_name(nic_id, create_config_file=True, fast=False):
    iface = None
    ifaces_path = '/sys/class/net'

    def look_for_nic():  # Helper function iterating over directory and looking for NIC on desired USB port (e.g. returns eth1 or wwan0)
        for filename in os.listdir(ifaces_path):
            path = os.readlink(ifaces_path + "/" + filename)
            if nic_id in path:
                return filename
        return None

    if ":" in nic_id:  # Get nic name from USB id
        for i in range(3):
            iface = look_for_nic()
            if iface:
                break
            if fast is False:
                sleep(3)
    else:
        with open(nic_id, "r") as f:
            iface = f.read()

    if not iface:
        return None

    if create_config_file:
        check_iface_config_file(iface)

    return iface


# Get LTE NIC name and creates appropriate config file if necessary


def remove_iface_confile(iface_name):
    if iface_name is None:
        return

    iface_path = get_lte_interface_confile_path(iface_name)

    if os.path.exists(iface_path):

        NetworkManagerDriver.set_down(iface_name, force=True)
        print("Removing network interface " + iface_name)
        os.remove(iface_path)
    else:
        print("No interface config found, nothing to remove...")


# Delete file previously created by check_iface_config_file. Used in purge/remove process.


def was_confile_changed():

    changed = False
    with open(CONFILE_PATH, "r+") as f:
        data = f.read().splitlines()
        m = hashlib.md5()
        for i in range(0, len(data) - 1): m.update(data[i].encode())
        if data[-1].strip(" #;\r\n\t") != m.hexdigest():
            changed = True
            signature = "; " + m.hexdigest()

            if ';' in data[-1][0:3]:
                data[-1] = signature
            else:  # Create new signature - used during migration from non-signed confile
                data.append("; Config file signature - do not modify it nor add any lines or whitespaces!")
                data.append(signature)

            f.seek(0)
            f.write("\n".join(data))

    return changed


# Check whether the config file has been changed since the last signature
# If so, signtarure is re-generated (also in case the signature not found)


def get_from_confile(config_parser, section, option, default, boolean=False, integer=False) -> str | int | bool:
    try:
        value = config_parser.get(section, option)
        if boolean:
            value = value.lower() in ("yes", "true", "t", "1")
        elif integer:
            value = int(value)
    except ValueError:
        logging.warning(
                "Cannot parse integer parameter {}:{}, using default value: {}".format(section, option, default))
        return default
    except:
        logging.debug("Cannot parse parameter {}:{}, using default value: {}".format(section, option, default))
        return default

    return value


class NetworkManagerDriver(object):
    class NMType(Enum):
        INTERFACES = 1
        SYSTEMD = 2

    Color = Enum('mmtype', ['networking', 'systemd'])

    @staticmethod
    def __set_up_networking(iface):
        subprocess.call("/sbin/ip link set dev {} up".format(iface), shell=True)
        subprocess.call("/sbin/ifup " + str(iface), shell=True)  # Can take longer

    @staticmethod
    def __set_down_networking(iface, force):
        subprocess.call("/sbin/ip link set dev {} down".format(iface), shell=True)
        subprocess.call("/sbin/ifdown {} ".format("--force " if force is True else "") + str(iface),
                        shell=True)  # Can take longer

    @staticmethod
    def __set_up_systemd(iface):
        subprocess.call("/sbin/ip link set dev {} up".format(iface), shell=True)
        subprocess.call("/bin/networkctl up " + iface, shell=True)

    @staticmethod
    def __set_down_systemd(iface, force):
        subprocess.call("/sbin/ip link set dev {} down".format(iface), shell=True)
        subprocess.call("/bin/networkctl down " + iface, shell=True)
        if force is True:
            subprocess.call("/bin/networkctl reload", shell=True)

    __up_function: callable = None
    __down_function: callable = None

    @classmethod
    def set_up(cls, iface):
        cls.__up_function(iface)

    @classmethod
    def set_down(cls, iface, force=False):
        cls.__down_function(iface, force)

    @classmethod
    def init(cls):
        if NetworkManagerDriver.get_network_manager_type() == NetworkManagerDriver.NMType.INTERFACES:
            cls.__up_function = NetworkManagerDriver.__set_up_networking
            cls.__down_function = NetworkManagerDriver.__set_down_networking
        elif NetworkManagerDriver.get_network_manager_type() == NetworkManagerDriver.NMType.SYSTEMD:
            cls.__up_function = NetworkManagerDriver.__set_up_systemd
            cls.__down_function = NetworkManagerDriver.__set_down_systemd

    @staticmethod
    def get_network_manager_type():
        if os.path.isfile(NIC_CONF_PATH_INTERFACES + "interfaces"):
            return NetworkManagerDriver.NMType.INTERFACES
        else:
            return NetworkManagerDriver.NMType.SYSTEMD


def pretty_class_name(obj):
    class_name = obj.__class__.__name__
    module_name = obj.__class__.__module__
    if module_name == "__main__":
        return class_name
    return f"{module_name}.{class_name}"


# Collects and computes statistical data related to cellular network quality and WAN network availability
class Stats(object):
    SYSFS_PATH = "/run/unipi_lte/"

    SYSFS_RSSI_FILE = "rssi"
    SYSFS_SIGQUAL_FILE = "sigqual"
    SYSFS_MODE_FILE = "mode"
    SYSFS_NETTYPE_FILE = "nettype"

    sysfs_files = (SYSFS_RSSI_FILE, SYSFS_SIGQUAL_FILE, SYSFS_MODE_FILE, SYSFS_NETTYPE_FILE)

    def __init__(self, period=120, ping_active=False):
        self.period = period
        self.states = set()
        self.ping_active = ping_active
        self.reset()

        if not os.path.exists(Stats.SYSFS_PATH):
            os.mkdir(Stats.SYSFS_PATH)

        for filename in Stats.sysfs_files:
            with open(Stats.SYSFS_PATH + filename, "w") as file:
                file.write("-1")

    @staticmethod
    def set_default_values():
        for filename in Stats.sysfs_files:
            with open(Stats.SYSFS_PATH + filename, "w") as file:
                file.write("-1")

    def __del__(self):
        Stats.set_default_values()

    def report_partial(self, rssi, sigqual, nettype):  # Commit often read values only, called from main loop

        try:
            rssi = int(rssi)
            sigqual = int(sigqual)
        except:
            return

        self.nettypes.add(nettype)

        if rssi == 99:  # Not valid value
            return

        self.cnt = self.cnt + 1
        if self.rssi_max < rssi:
            self.rssi_max = rssi
        elif self.rssi_min > rssi:
            self.rssi_min = rssi
        self.rssi_sum += rssi

        self.sigqual = sigqual

        with open(stats.SYSFS_PATH + Stats.SYSFS_NETTYPE_FILE, "w") as file:
            file.write(str(nettype))

        # Mirror values to sysfs

        with open(stats.SYSFS_PATH + Stats.SYSFS_RSSI_FILE, "w") as file:
            file.write(str(rssi))

        with open(stats.SYSFS_PATH + Stats.SYSFS_SIGQUAL_FILE, "w") as file:
            file.write(str(sigqual))

    def report_metrics(self, rssi, sigqual, wan_status, nettype,
                       ping_result=None):  # Commit all values, called form main loop

        self.report_partial(rssi, sigqual, nettype)

        self.states.add(wan_status)

        if ping_result is not None:
            if ping_result is not True:
                self.pings_failed += 1

            self.pings_cnt += 1

        with open(stats.SYSFS_PATH + Stats.SYSFS_MODE_FILE, "w") as file:
            file.write(str(wan_status))

    def reset(self):
        self.cnt = 0
        self.rssi_sum = 0
        self.rssi_min = 100
        self.rssi_max = 0
        self.pings_cnt = 0
        self.pings_failed = 0
        self.last_summary = time.time()
        self.sigqual = 0
        self.nettypes = set()
        self.states = set()

    def get_summmary(self, do_reset=True):  # Get report output in text form (e.g. for periodical storing to a log file)

        if self.ping_active and self.pings_cnt != 0:
            sum_str = "RSSI: (MIN {}, MAX {}, AVG {}), SIGQUAL: {}, NET-TYPES: {}, USED_MODES: {} PING_COUNT: {} PING_SUCCESS_RATIO: {:0.2f}".format(
                self.rssi_min, self.rssi_max, int(self.rssi_sum / self.cnt), self.sigqual, list(self.nettypes),
                self.states, self.pings_cnt, ((self.pings_cnt - self.pings_failed) / self.pings_cnt) * 100)
        elif self.cnt > 0:
            sum_str = "RSSI: (MIN {}, MAX {}, AVG {}) SIGQUAL: {} NET-TYPES: {} ".format(self.rssi_min, self.rssi_max,
                int(self.rssi_sum / self.cnt), self.sigqual, list(self.nettypes))
        else:
            sum_str = "Not enough samples to compute stats"

        if do_reset: self.reset()
        return sum_str

    def is_time_fullfilled(self):  # Returns true if it is time to generate stats into the log file
        if not self.period:
            return False
        return (self.last_summary + self.period) < time.time()


# Provides ping and autoswitch functionality
class ConnectionGuard(object):
    GW_MASK_LTE = 1
    GW_MASK_ETH = 2

    def isOnline(self):  # True if the unit has access to the WAN network (found by ping)

        if self.wan_state not in (WAN_STATE_LTE_OFF, WAN_STATE_LTE_OFF_ETH_OFF, WAN_STATE_INIT):
            return False
        return True

    def __init__(self):

        global current_config
        self.local_lte_ip = None
        self.current_eth_metric = 0
        self.current_lte_metric = 0
        self.gw_bitmask = 0
        self.rtable_valid = True
        self.autoswitch_was_active = False
        self.reset_state()
        self.ping_executor = PingMan(ping_timeout=current_config['ping']['timeout'],
                                     ping_target=current_config['ping']['target_lte'])

    def reset_state(self):
        self.wan_state = WAN_STATE_INIT if current_config['ping']['active'] is True else WAN_STATE_MONITOR_OFF
        self.fail_counterA = 0
        self.fail_counterB = 0
        self.rescue_counter = 0

    # Detects if the main interface is not available (e.g. cable is off)
    # Reads and compares metrics of the intefaces (both main and backup)
    def __read_default_gateway(self):
        global current_config

        self.current_eth_metric = -1
        self.current_lte_metric = -1
        gw_bitmask = 0
        routecount = 0

        STATE_PATH = "/sys/class/net/{}/operstate"
        upstrings = ("up", "unknown")

        try:
            with open(STATE_PATH.format(current_config['ping']['eth_iface']), "r") as f:
                res = f.read()
                if any(x in res for x in upstrings):
                    gw_bitmask |= self.GW_MASK_ETH
        except:
            pass

        try:
            with open(STATE_PATH.format(current_config['lte_iface']), "r") as f:
                res = f.read()
                if any(x in res for x in upstrings):
                    gw_bitmask |= self.GW_MASK_LTE
        except:
            pass

        with open("/proc/net/route") as fh:
            for line in fh:
                routecount = routecount + 1
                fields = line.strip().split()  # Default gateway ha
                if (fields[1] == '00000000') and (fields[0] == current_config['ping']['eth_iface']):
                    # If more than one route for main interface exists, use the lowest metric
                    if self.current_eth_metric == -1:
                        self.current_eth_metric = int(fields[6])
                    elif int(fields[6]) < self.current_eth_metric:
                        self.current_eth_metric = int(fields[6])
                elif (fields[1] == '00000000') and (fields[0] == current_config['lte_iface']):
                    self.current_lte_metric = int(fields[6])

                    if self.local_lte_ip is None:
                        buffer = bytearray(4)
                        struct.pack_into("<L", buffer, 0, int(fields[2], 16))
                        self.local_lte_ip = ".".join(str(c) for c in buffer)
                        logging.debug("Assigned local IP " + self.local_lte_ip)

                        if self.local_lte_ip == "0.0.0.0":
                            self.local_lte_ip = None  # logging.debug("GW IP {}".format(self.local_lte_ip))

        if self.current_eth_metric == -1:
            logging.debug("Cannot obtain metric for {} (is up: {})".format(current_config['ping']['eth_iface'],
                                                                          gw_bitmask & self.GW_MASK_ETH))
            gw_bitmask &= ~(self.GW_MASK_ETH)

        if self.current_lte_metric == -1:
            logging.debug("Cannot obtain metric for {} (is up: {})".format(current_config['lte_iface'],
                                                                          gw_bitmask & self.GW_MASK_LTE))
            gw_bitmask &= ~(self.GW_MASK_LTE)

        is_rtable_valid = (True if (routecount > 2) else False)
        # is_current_iface_main = (True if ((main_metric < backup_metric) and (gw_bitmask & 1)) else False)

        logging.debug(
            "Current iface metrics main_iface:{} backup_iface:{} Local LTE ip {}, routing items {}, gw_bitmask {}".format(
                self.current_eth_metric, self.current_lte_metric, self.local_lte_ip, routecount, gw_bitmask))

        return (gw_bitmask, is_rtable_valid)

    def keep_iface_on(self):

        if current_config['ping']['autoswitch'] is False:
            self.gw_bitmask, is_rtable_valid = self.__read_default_gateway()

        if (not (self.gw_bitmask & self.GW_MASK_LTE)):
            logging.warning(
                "UP {} LTE default route is MISSING, restart the iface...".format(current_config['lte_iface']))
            if self.try_interface_up(current_config['lte_iface'], down_firstly=True) is False:
                raise ModemConnectionError()

    @staticmethod
    def try_interface_up(iface, down_firstly=False):
        if down_firstly is True:
            NetworkManagerDriver.set_down(iface, force=True)

        def has_ip_assigned():
            ret = subprocess.check_output("ip addr show " + iface, shell=True, text=True)
            logging.debug(ret)
            if "inet " in ret and "DOWN" not in ret and "169.254" not in ret:
                return True
            return False

        if has_ip_assigned() is True and down_firstly is False:
            return True

        for i in range(2):  # Wait max 18 sec for DHCP IP address assignment
            NetworkManagerDriver.set_up(iface)
            for j in range(8):
                if has_ip_assigned() is True: return
                logging.debug(
                    "Waiting for IP assign...remain {}:{} down firstly {}".format(str(2 - i), str(8 - j), down_firstly))
                time.sleep(8)

        if down_firstly is False:
            logging.warning("Cannot obtain IP address from LTEs DHCP server...iface restart")
            return ConnectionGuard.try_interface_up(iface, down_firstly=True)

        else:
            logging.error("Cannot obtain IP address from LTEs DHCP server...")
            return False

    @staticmethod
    def update_resolver_file(lte_as_dns, dns_ip=None):


        if NetworkManagerDriver.get_network_manager_type() == NetworkManagerDriver.NMType.SYSTEMD:
            if lte_as_dns is True:
                logging.debug("Changing domain for DNS resolving to LTE iface ")
                ret = subprocess.check_output("resolvectl domain {} ~.".format(current_config['lte_iface']), shell=True)
            else:
                logging.debug("Changing domain for DNS resolving to main iface ")
                ret = subprocess.check_output("resolvectl domain {} \"\"".format(current_config['lte_iface']),
                                              shell=True)

            return

        FILEPATH = "/etc/resolvconf/resolv.conf.d/head"
        LINE = "nameserver {}\n".format(dns_ip)  # Toto vznikne
        data = None
        changed = False

        def remove_current_records(data):
            data = str(data).splitlines(True)
            new_content = ""
            line_num = 0
            changed = False
            while line_num < len(data):
                if "unipi-lte" in data[line_num]:  # skip
                    line_num = line_num + 2
                    changed = True
                    continue
                new_content += (data[line_num])
                line_num = line_num + 1

            with open(FILEPATH, "w") as f:
                f.write(new_content)

            return changed

        try:  # If the config file does not exists
            with open(FILEPATH, "r") as f:
                data = f.read()
        except Exception as e:
            logging.warning("Cannot change DNS resolver file: " + str(e))
            return

        if LINE in data:
            newdata = re.sub(r"#?({})".format(LINE), r"{}\1".format("" if lte_as_dns is True else "#"), data)
            if newdata != data:
                with open(FILEPATH, "w") as f:
                    f.write(newdata)
                changed = True

        elif lte_as_dns == True and dns_ip is not None:
            remove_current_records(data)
            MSG = "#This line has been added by unipi-lte daemon\n{}".format(LINE)
            with open(FILEPATH, "a") as f:
                f.write(MSG)
            changed = True

        else:  # Remove records
            changed = remove_current_records(data)

        if changed is not False:
            status = subprocess.call("resolvconf " + "-u", shell=True)

    def __switch_default_iface(self, toBackup):  # Switches network ifaces (change their metrics)
        global current_config

        new_metric_shift = 0
        # DEF Change metric - lokalni funkce

        if toBackup and (self.gw_bitmask & self.GW_MASK_LTE):

            logging.debug("Switching to LTE interface {}".format(current_config['lte_iface']))

            if not (self.gw_bitmask & self.GW_MASK_LTE):
                return

            if self.current_lte_metric + 1 == self.current_eth_metric:
                new_metric_shift = 1

            new_metric = self.current_eth_metric - 1 - new_metric_shift
            if new_metric <= 0:
                new_metric = 1

            logging.debug(
                "ip route add default via {} dev {} metric {}".format(self.local_lte_ip, current_config['lte_iface'],
                                                                      new_metric))
            logging.debug(
                "ip route delete default dev {} metric {}".format(current_config['lte_iface'], self.current_lte_metric))

            os.system(
                "ip route add default via {} dev {} metric {}".format(self.local_lte_ip, current_config['lte_iface'],
                                                                      new_metric))
            os.system(
                "ip route delete default dev {} metric {}".format(current_config['lte_iface'], self.current_lte_metric))

            self.current_lte_metric = new_metric

            try:
                self.update_resolver_file(lte_as_dns=True, dns_ip=self.local_lte_ip)
            except Exception as E:
                logging.critical("Failure: " + str(E))

        elif self.gw_bitmask & self.GW_MASK_ETH:

            logging.debug("Switching to main interface {}".format(current_config['ping']['eth_iface']))

            if self.current_lte_metric - 1 == self.current_eth_metric:
                new_metric_shift = 1

            new_metric = self.current_eth_metric + 1 + new_metric_shift

            logging.debug(
                "ip route add default via {} dev {} metric {}".format(self.local_lte_ip, current_config['lte_iface'],
                                                                      new_metric))
            logging.debug(
                "ip route delete default dev {} metric {}".format(current_config['lte_iface'], self.current_lte_metric))
            os.system(
                "ip route add default via {} dev {} metric {}".format(self.local_lte_ip, current_config['lte_iface'],
                                                                      new_metric))
            os.system(
                "ip route delete default dev {} metric {}".format(current_config['lte_iface'], self.current_lte_metric))

            self.current_lte_metric = new_metric

            try:
                self.update_resolver_file(lte_as_dns=False, dns_ip=self.local_lte_ip)
            except Exception as E:
                logging.critical("Failure: " + str(E))

        logging.debug("Autoswitch complete")

    def get_wan_status(self):  # Get current status of the connection (see WAN_STATE_xxxx constants)
        return self.wan_state

    def get_rtable_valid(self):  # Returns true if routing table contains at least one valid record
        return self.rtable_valid

    def __autoswitch(self, lte_status):  # Autoswitch core called from ping function (if enabled in the config file)

        global current_config

        LTE_UNCHANGED = 0
        LTE_OK = 1
        LTE_OFFLINE = 2

        ETH_UNCAHNGED = 0
        ETH_GOES_OFFLINE = 1
        ETH_GOES_ONLINE = 2
        eth_status = ETH_UNCAHNGED

        ping_eth_ok: bool = False  # Ping failed as default

        self.gw_bitmask, self.rtable_valid = self.__read_default_gateway()

        logging.debug("GW MASK: " + str(self.gw_bitmask))

        if not self.rtable_valid:  # Do not perform autoswitch if routing table is not valid (typically after startup)
            logging.debug("Routing table is not (yet) valid, autoswitch skipping...")
            return

        elif self.gw_bitmask != (self.GW_MASK_ETH | self.GW_MASK_LTE) and self.autoswitch_was_active is False:
            logging.debug("No more default gateways, autoswitch skipping...")
            if lte_status == LTE_OK:
                if self.wan_state != WAN_STATE_LTE_ON_ETH_OFF:
                    self.__switch_default_iface(toBackup=True)
                self.wan_state = WAN_STATE_LTE_ON_ETH_OFF
            else:
                self.wan_state = WAN_STATE_LTE_OFF_ETH_OFF
            return

        else:
            self.autoswitch_was_active = True
            logging.debug("Autoswitch ACTIVE ")

        # -----------------------------------------------------------------------------------------------------------

        if self.gw_bitmask & self.GW_MASK_ETH:  # If the main iface is present, ping it
            ping_eth_ok = self.ping_executor.do_ping(current_config['ping']['eth_iface'])

        elif self.wan_state in {WAN_STATE_LTE_OFF_ETH_ON, WAN_STATE_LTE_ON_ETH_ON,
                                WAN_STATE_INIT}:  # If main interface has been disconnected, switch to backup immediately
            self.fail_counterB = current_config['ping']['threshold'] + 1
            logging.info("Main GW is not present ! Switching to backup iface {} immediately...".format(
                current_config['lte_iface']))

        if ping_eth_ok is False and self.wan_state not in {WAN_STATE_ETH_OFF, WAN_STATE_LTE_OFF_ETH_OFF,
                                                           WAN_STATE_LTE_ON_ETH_OFF}:  # Ping on main interface failed and ETH is ON

            if self.fail_counterB >= current_config['ping']['threshold']:  # If counter overflows, ETH ON -> ETH OFF
                eth_status = ETH_GOES_OFFLINE
                self.__switch_default_iface(toBackup=True)  # Switch to backup
                self.gw_bitmask, self.rtable_valid = self.__read_default_gateway()  # Refresh GW info
                if self.fail_counterB == current_config['ping']['threshold']:
                    logging.info("Ping failure threshold reached {} on main {}.Switching to backup {}...".format(
                        current_config['ping']['threshold'], current_config['ping']['eth_iface'],
                        current_config['lte_iface']))

            else:
                self.fail_counterB = self.fail_counterB + 1

            logging.debug("Ping failed on {} no. {}; Threshold: {}".format(current_config['ping']['eth_iface'],
                self.fail_counterB, current_config['ping']['threshold']))

        elif ping_eth_ok and self.wan_state not in {WAN_STATE_ETH_ON, WAN_STATE_LTE_OFF_ETH_ON,
                                                    WAN_STATE_LTE_ON_ETH_ON}:  # Ping OK, switch back to main iface

            if self.wan_state != WAN_STATE_INIT:
                logging.info(
                    "Main interface {} becomes available, switching to it".format(current_config['ping']['eth_iface']))
                self.__switch_default_iface(toBackup=False)  # Switch to main
                self.gw_bitmask, self.rtable_valid = self.__read_default_gateway()  # Refresh GW info

            eth_status = ETH_GOES_ONLINE
            logging.debug("ETH Ping OK")

        else:
            if ping_eth_ok:
                logging.debug("ETH Ping OK")
            self.fail_counterB = 0

        for i in range(2):  # Resolve current state variable
            if eth_status == ETH_GOES_ONLINE:
                if self.wan_state in (WAN_STATE_INIT, WAN_STATE_ETH_OFF):
                    self.wan_state = WAN_STATE_ETH_ON
                elif self.wan_state in (WAN_STATE_LTE_ON, WAN_STATE_LTE_ON_ETH_OFF):
                    self.wan_state = WAN_STATE_LTE_ON_ETH_ON
                elif self.wan_state in (WAN_STATE_LTE_OFF, WAN_STATE_LTE_OFF_ETH_OFF):
                    self.wan_state = WAN_STATE_LTE_OFF_ETH_ON

            elif eth_status == ETH_GOES_OFFLINE:
                if self.wan_state in (WAN_STATE_INIT, WAN_STATE_ETH_ON):
                    self.wan_state = WAN_STATE_ETH_OFF
                elif self.wan_state in (WAN_STATE_LTE_ON, WAN_STATE_LTE_ON_ETH_ON):
                    self.wan_state = WAN_STATE_LTE_ON_ETH_OFF
                elif self.wan_state in (WAN_STATE_LTE_OFF, WAN_STATE_LTE_OFF_ETH_ON):
                    self.wan_state = WAN_STATE_LTE_OFF_ETH_OFF

            if lte_status == LTE_OK:
                if self.wan_state in (WAN_STATE_INIT, WAN_STATE_LTE_OFF):
                    self.wan_state = WAN_STATE_LTE_ON
                elif self.wan_state in (WAN_STATE_ETH_ON, WAN_STATE_LTE_OFF_ETH_ON):
                    self.wan_state = WAN_STATE_LTE_ON_ETH_ON
                elif self.wan_state in (WAN_STATE_ETH_OFF, WAN_STATE_LTE_OFF_ETH_OFF):
                    self.wan_state = WAN_STATE_LTE_ON_ETH_OFF

            elif lte_status == LTE_OFFLINE:
                if self.wan_state in (WAN_STATE_INIT, WAN_STATE_LTE_ON):
                    self.wan_state = WAN_STATE_LTE_OFF
                elif self.wan_state in (WAN_STATE_ETH_ON, WAN_STATE_LTE_ON_ETH_ON):
                    self.wan_state = WAN_STATE_LTE_OFF_ETH_ON
                elif self.wan_state in (WAN_STATE_ETH_OFF, WAN_STATE_LTE_ON_ETH_OFF):
                    self.wan_state = WAN_STATE_LTE_OFF_ETH_OFF

        if self.gw_bitmask == (self.GW_MASK_ETH | self.GW_MASK_LTE):

            if (self.wan_state == WAN_STATE_LTE_ON_ETH_OFF) and (
                    self.current_lte_metric >= self.current_eth_metric):  # In backup mode
                logging.warning(
                    "Routing table metric mismatch A (changed by other application?), re-setting... {} {}".format(
                        self.current_lte_metric, self.current_eth_metric))
                self.__switch_default_iface(toBackup=True)

            elif (self.wan_state in (WAN_STATE_LTE_OFF_ETH_ON, WAN_STATE_LTE_ON_ETH_ON)) and (
                    self.current_eth_metric >= self.current_lte_metric):
                logging.warning("Routing table metric mismatch B (changed by other application?), re-setting...")
                self.__switch_default_iface(toBackup=False)

        return True  # return not (bool(response_main and response_backup))

    def ping(self):  # Checks the WAN connection, set state varaibles(s) and handle autoswitch stuff if enabled
        global current_config

        LTE_UNCHANGED = 0
        LTE_OK = 1
        LTE_OFFLINE = 2
        lte_status = LTE_UNCHANGED

        # Ping on the backup iface (LTE)
        ping_lte_ok = self.ping_executor.do_ping(current_config['lte_iface'])

        if ping_lte_ok is False:  # Ping not successful
            if self.rescue_counter >= 0:
                logging.debug("RESCUE-current: {}\nRESCUE-max: {}".format(self.rescue_counter, RESCUE_THRESHOLD))
                if self.rescue_counter > RESCUE_THRESHOLD:
                    self.rescue_counter = 0
                    raise ModemConnectionError

                self.rescue_counter = self.rescue_counter + 1

            logging.debug(
                "Ping failed GSM no. {}; Threshold: {}".format(self.fail_counterA, current_config['ping']['threshold']))
            if self.fail_counterA >= current_config['ping']['threshold']:

                if self.wan_state in {WAN_STATE_INIT, WAN_STATE_ETH_ON}:
                    logging.info("LTE: Cannot get to online state {}".format(self.wan_state))
                elif self.wan_state in {WAN_STATE_LTE_ON, WAN_STATE_LTE_ON_ETH_OFF, WAN_STATE_LTE_ON_ETH_ON}:
                    logging.info("LTE: Goes offline (previously online)")

                lte_status = LTE_OFFLINE
            else:
                self.fail_counterA = self.fail_counterA + 1
        else:  # Ping was OK
            self.fail_counterA = 0
            self.rescue_counter = 0
            if self.wan_state in {WAN_STATE_LTE_OFF, WAN_STATE_LTE_OFF_ETH_OFF, WAN_STATE_LTE_OFF_ETH_ON}:
                logging.info("LTE: Goes online (previously offline) {}".format(self.wan_state))
            lte_status = LTE_OK
            logging.debug("LTE Ping OK")

        if current_config['ping']['autoswitch'] is True:  # Autoswitch feature is active and LTE status is resolved
            self.__autoswitch(lte_status)
        else:  # Autoswitch disabled, only backup iface periodical probing is active
            if lte_status == LTE_OK:
                self.wan_state = WAN_STATE_LTE_ON
            elif lte_status == LTE_OFFLINE:
                self.wan_state = WAN_STATE_LTE_OFF

        logging.debug("Current WAN state {}".format(self.wan_state))

        return True if lte_status == LTE_OK else False


def do_action(modem, action,
              errcode=ERROR_CODE_UNSPEC):  # Perform desired action (as a reaction to different error conditions)

    reinit = False

    if action == ACTION_SOFTRESTART:
        logging.info("Modem soft-restart")
        try:
            modem.restart(soft=True)
        except Exception as e:
            logging.info("Modem soft-restart failed, performing HARD-restart: " + str(e))
            do_action(modem, ACTION_HARDRESTART)  # Hard restart as a fallback to the SOFT one (via AT command)

        reinit = True  # Re-init modem

    elif action == ACTION_HARDRESTART:
        logging.info("Modem hard-restart caused by an error")
        NetworkManagerDriver.set_down(current_config['lte_iface'], force=True)

        if modem:
            modem.restart_hw()
            del modem

        reinit = True  # Re-init modem

    elif action == ACTION_SUSPEND:
        logging.warning("Daemon suspended...exit in {} sec".format(current_config['suspend_timeout']))
        if current_config['suspend_timeout'] < 0:
            do_action(modem, ACTION_HALT, errcode)
        if modem:
            modem.go_halt_suspend(errcode)
        time.sleep(current_config['suspend_timeout'])
        reinit = True  # Re-init modem

    elif action == ACTION_HALT:
        logging.warning("Daemon halted forever...must be restarted manually {}".format(errcode))
        if modem:
            modem.go_halt_suspend(errcode)
        while True:
            time.sleep(current_config['suspend_timeout'])

    return reinit


class ModemFactory(object):

    # Try to read current unipi platform based on CPU type
    # TODO - use unipi-id (always present since Debian 11) instead
    @staticmethod
    def get_platform():
        model = "Generic"
        family = "N/A"

        # Primary (main) way
        UNIPI_ID_MODEL = "/run/unipi-plc/unipi-id/product_model"
        UNIPI_ID_FAMILY = "/run/unipi-plc/unipi-id/product_family"
        UNIPI_ID_DESCRIPTION = "/run/unipi-plc/unipi-id/product_description"

        # Secondary (backup) way
        MODEL_ID_FILE = "/sys/firmware/devicetree/base/unipi-model"
        PROCINFO_FILE = "/proc/cpuinfo"

        if os.path.isfile(UNIPI_ID_MODEL) and os.path.isfile(UNIPI_ID_FAMILY):
            # use unipi-id informations
            with open(UNIPI_ID_MODEL, 'r') as f:
                model = f.read().strip()
            with open(UNIPI_ID_FAMILY, 'r') as f:
                family = f.read().strip()
        elif os.path.isfile(UNIPI_ID_DESCRIPTION):
            with open(UNIPI_ID_DESCRIPTION, 'r') as f:
                data = f.read().strip().splitlines()
                for line in data:
                    if line.startswith("Platform family"):
                        family = line.split(':')[1].split(' ')[1].strip()
                    elif line.startswith("Product model"):
                        model = line.split(':')[1].strip()

        # Backup way
        elif os.path.isfile(MODEL_ID_FILE):
            with open(MODEL_ID_FILE, 'rb') as f:
                model = f.read()
                printables = {ord(c) for c in string.printable}
                model = "".join(chr(c) for c in model if c in printables)
            if len(model) == 4 and model[3] == '7':
                family = "Patron"
            elif len(model) == 4 and model[3] == '5':
                family = "Axon"

        return family, model

    @staticmethod
    def modem_discovery(supported_modems: dict[tuple[int, int], str], pwr_control: LTEModemPWRControl):

        modem_found = None
        basepath = '/sys/bus/usb/devices/'

        ATTEMPTS_PRESET = 10
        attempts = ATTEMPTS_PRESET

        while attempts > 0:
            logging.debug(f"Looking for modem attempt {attempts}")
            for device in os.listdir(basepath):
                device_path = os.path.join(basepath, device)

                # Check if the VID and PID files exist in the directory
                vid_path = os.path.join(device_path, 'idVendor')
                pid_path = os.path.join(device_path, 'idProduct')

                if os.path.exists(vid_path) and os.path.exists(pid_path):
                    # Read the VID and PID values
                    with open(vid_path, 'r') as vid_file:
                        vid = int(vid_file.read().strip(), 16)

                    with open(pid_path, 'r') as pid_file:
                        pid = int(pid_file.read().strip(), 16)

                    # Take the first matching item in the list
                    if (vid, pid) in supported_modems:
                        sleep(1)
                        return USBModemDevice(supported_modems[(vid, pid)], vid,
                                              pid)  # modem_found = next((_modem for _modem in supported_modems if (_modem.vid == vid and _modem.pid == pid)), None)

            attempts -= 1
            if attempts < (ATTEMPTS_PRESET / 4):
                logging.debug(f"HW restart during discovery")
                pwr_control.hw_restart()
                sleep(10)
            elif attempts < (ATTEMPTS_PRESET / 2):
                logging.debug(f"Modem enable during discovery")
                pwr_control.enable()
                sleep(5)
            else:
                sleep(10 - attempts)

        return None

    @staticmethod
    def create_modem_instance() -> LTEModemPlatform | None:
        # Resolve platform

        platform, model = ModemFactory.get_platform()

        logging.info("Running on {} series, model: {} ".format(platform, model))

        match platform:
            case "Iris" | "IRIS":
                leds = LedServer(PLATFORM_SPEC["ID-LTE-1"]['leds'])
                leds.all_blink()

            case "Patron":
                leds = LedServer(DEFAULT_LED_PATHS)
                leds.all_blink()

            case "Edge" | "Cm40":
                leds = LedServer(DEFAULT_LED_PATHS)
            case _:
                logging.error("Running on unsupported {} series, model: {}, exiting.".format(platform, model))
                return None

        power_ctrl = LTEModemPWRControl()
        power_ctrl.enable()

        new_usb_device = ModemFactory.modem_discovery(supported_modems=SUPPORTED_USB_MODEMS,
                                                      pwr_control=power_ctrl)
        if new_usb_device is None:
            logging.error("Modem not found (slot empty?)")
            return LTEModemPlatform(None, leds, power_ctrl)

        logging.info(f"Found USB device: {new_usb_device}")

        new_modem = None
        match new_usb_device.device_name:
            case "EC200U" | "EC200A" | "EG912Y" | "EG915U":
                new_modem = LTEModemQuectelGeneric(usb_device=new_usb_device)

        _modem_platform = LTEModemPlatform(new_modem, leds, power_ctrl)
        return _modem_platform



# --------------------------------------------------------------------------
# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

# Entrypoint is here!

# Global (but not super-global) variables
modem: LTEModemPlatform | None = None
ping_executor: PingMan | None = None


parser = argparse.ArgumentParser(description="LTE connection management daemon", epilog="unipi.technology 2019-2024")
parser.add_argument("-r", "--reset", dest="reset", action="store_true", help="perform HW reset of the LTE module")
parser.add_argument("-d", "--delete", dest="delete", action="store_true", help="delete NIC from the system")
parser.add_argument("-c", "--console", dest="console", action="store_true", help="write logs to console")
parser.add_argument("-e", "--debug", dest="debug", action="store_true", help="increase log level to debug")
args = parser.parse_args()

LogMaster.logger_init(args.console)

NetworkManagerDriver.init()
stats = None

if args.reset:  # Single shot reset (call the script with -r or --reset)
    _modem_platform = ModemFactory.create_modem_instance()
    _modem_platform.restart_hw()
    _modem_platform.do_cleanup()
    sys.exit(0)

if os.path.isfile(PIDFILE_PATH):
    print("LTE Daemon already running, exiting...")
    sys.exit(0)

if args.delete:  # Single shot reset (call the script with -r or --reset)
    _modem_platform: LTEModemPlatform = ModemFactory.create_modem_instance()
    if not _modem_platform or \
       not _modem_platform.modem or \
       not _modem_platform.modem.nic_name:
        sys.exit(0)

    current_config['lte_iface'] = get_lte_iface_name(_modem_platform.modem.nic_name, create_config_file=False)
    remove_iface_confile(current_config['lte_iface'])
    _modem_platform.do_cleanup()

    ConnectionGuard.update_resolver_file(lte_as_dns=False, dns_ip=None)
    sys.exit(0)

try:  # Daemon mode starts here

    pid = str(os.getpid())
    file = open(PIDFILE_PATH, 'w')
    file.write(pid)
    file.close()

    signal.signal(signal.SIGINT, sig_handler)  # LOG exit cause by signal in handler
    signal.signal(signal.SIGTERM, sig_handler)
    signal.signal(signal.SIGQUIT, sig_handler)

    conn_quard = None
    try:
        config = configparser.ConfigParser(inline_comment_prefixes=(';',))
    except TypeError:
        config = configparser.ConfigParser()  # In Python 2.x, the module is slightly different

    err_msg = ""

    if not os.path.isfile(CONFILE_PATH):  # Check the config file
        err_msg = "Config file not found on path: {}".format(CONFILE_PATH)
        confile_was_changed = False
    else:
        confile_was_changed = was_confile_changed()

    try:
        config.read([CONFILE_PATH])
    except Exception as e:
        err_msg = "Error while parsing file {} : {}".format(CONFILE_PATH, str(e))

    log_path = get_from_confile(config, "LOG", "path", DEFAULT_LOG_PATH)
    log_level = get_from_confile(config, "LOG", "severity", DEFAULT_LOGGING_LEVEL).upper()
    if args.debug:
        log_level = 'DEBUG'

    if log_path is not None and not args.console:
        log_size = get_from_confile(config, "LOG", "logfile_size", DEFAULT_LOGFILE_SIZE, integer=True)
        LogMaster.logger_swich_to_file(log_path, log_size)
        #print("Log path set, logging to " + str(log_path))

    logging.info("running version {} from {} - Logging started..".format(DAEMON_VERSION, datetime.datetime.fromtimestamp(
        os.path.getmtime(os.path.realpath(__file__)))))

    logger = LogMaster.getLogger()
    try:
        logger.setLevel(log_level.upper())
    except ValueError as E:
        logger.setLevel(DEFAULT_LOGGING_LEVEL)
        err_msg = "Invalid log severity, see the doc."

    if err_msg:  # Handle log messages from config file parsing stage
        logging.warning(err_msg + "; USING DEFAULT VALUES")
    else:
        logging.debug("Config file {} parsed".format(CONFILE_PATH))

    current_config = {'_was_changed': confile_was_changed,
                      'apn': get_from_confile(config, "MAIN", "apn", None),
                      'conn_type': get_from_confile(config, "MAIN", "connection_type", None)}

    if current_config['apn'] is None:
        modem = ModemFactory.create_modem_instance()
        modem.leds.all_off()
        # os.system("ip link set {} down".format(current_config['lte_iface']))
        print("No APN defined in {}, stopping unipi-lte service...".format(CONFILE_PATH))
        logging.warning("No APN defined in {}, stopping unipi-lte service...".format(CONFILE_PATH))
        os._exit(0)

    if current_config['conn_type'] is None:
        _conn_type_apn = current_config['apn'].split(',')
        if len(_conn_type_apn) > 1:
            current_config['conn_type'] = _conn_type_apn[0]
            current_config['apn'] = _conn_type_apn[1]
            logging.warning("You are using old configuration option apn. Split it into apn and connection_type")
        else:
            current_config['conn_type'] = "IP"

    current_config['conn_type'] = current_config['conn_type'].strip('"')
    current_config['apn'] = current_config['apn'].strip('"')

    current_config['auth_method'] = get_from_confile(config, "MAIN", "auth_method", 0, integer=True)
    if current_config['auth_method'] != 0:
        current_config["auth_user"] = get_from_confile(config, "MAIN", "auth_user", "")
        current_config["auth_pass"] = get_from_confile(config, "MAIN", "auth_pass", "")

    current_config['pin'] = get_from_confile(config, "MAIN", "pin", None)
    current_config['usbmode'] = get_from_confile(config, "MAIN", "usbmode", "2")
    current_config['report_period'] = get_from_confile(config, "MAIN", "report_period", 86400, integer=True)
    current_config['oneshot'] = get_from_confile(config, "MAIN", "oneshot", default=False, boolean=True)
    current_config['regtimeout'] = get_from_confile(config, "MAIN", "reg_timeout", 120, integer=True)
    current_config['suspend_timeout'] = get_from_confile(config, "MAIN", "suspend_timeout", 86400, integer=True)
    current_config['lte_autokeep_on'] = get_from_confile(config, "MAIN", "auto_keepon", default=True, boolean=True)

    current_config['lte_iface'] = None

    current_config['ping'] = {}
    current_config['ping']['active'] = get_from_confile(config, "PING", "active", default=False, boolean=True)

    current_config['ping']['timeout'] = int(get_from_confile(config, "PING", "timeout", 60, integer=True))
    current_config['ping']['period'] = int(get_from_confile(config, "PING", "period", 20))
    current_config['ping']['threshold'] = int(get_from_confile(config, "PING", "fail_threshold", 3))
    current_config['ping']['target_lte'] = get_from_confile(config, "PING", "target_lte", "8.8.8.8")
    current_config['ping']['autoswitch'] = get_from_confile(config, "PING", "autoswitch", default=False, boolean=True)
    current_config['ping']['eth_iface'] = get_from_confile(config, "PING", "primary_iface", "eth0")

    if current_config['ping']['autoswitch']:
        current_config['ping']['target_eth'] = get_from_confile(config, "PING", "target_eth", "8.8.4.4")

    conn_quard = ConnectionGuard()

    logging.debug("CURRENT SETTINGS: {} ".format(str(current_config)))

    # Regex parsing some specific AT commands replies

    REGEX_IP = re.compile(r"\+UIPADDR:\s*\d+,\"\w*\",\"([^\"]*)\"")
    REGEX_PRIV_IP = re.compile(r"\+UIPCONF:\s\"([^\"]*)\",\"([^\"]*)\"")

    # In-loop variables

    stats = Stats(current_config['report_period'], current_config['ping']['active'])
    logging.debug("Entering the main loop...")
    ping_timestep = 0
    reg_timeout = current_config['regtimeout']
    is_registered = False
    start_timestamp = 0

    reinit = True  # Init after start
    modem = None

    while True:
        try:
            start_timestamp = time.time()
            if reinit:

                if conn_quard is not None:
                    conn_quard.reset_state()

                reinit = False
                wan_available = False
                # if not modem: # Try to create modem instance - working serial port needed
                if modem is not None:
                    modem.do_cleanup()

                modem = ModemFactory.create_modem_instance()

                if modem is None or modem.modem is None:
                    logging.error("Exitting...")
                    sys.exit(0)

                logging.debug(f"Created instance of modem type {pretty_class_name(modem)}")

                modem.init()

                ping_timestep = 0
                reg_timeout = current_config['regtimeout']
                is_registered = False
                start_timestamp = 0

                # ------------------------------------------------------------------------------

                if current_config['oneshot']:
                    logging.info("One-shot mode, exiting...")
                    sys.exit(0)

            rssi, sigqual = modem.get_sig_metrics()

            operator, nettype = modem.get_operator_nettype()

            # logging.critical(f"Schlaffen........--- {rssi} {sigqual} {operator} {nettype}")
            # sleep(569)

            if operator is not None:  # Registered to the network, thus nettype is valid

                reg_timeout = current_config['regtimeout']  # Reset non-registered counter

                if is_registered is False:  # Registered for a first time
                    is_registered = True
                    logging.info(
                        "Registered to: {} Network type: {} RSSI: {} Sigqual: {}".format(operator, nettype, rssi,
                                                                                         sigqual))

                    # Check local IP address assigned by modems DHCP server

                    ip, priv_ip = modem.get_ip_addresses()
                    logging.info("WAN IP ADDRESS: {} INTERFACE IP ADDRESS: {}".format(ip, priv_ip))

                    if ip is not None:
                        """
                        # TODO - only for modems that supports it
                        new_net_settings = resolve_subnets(ip.group(1), priv_ip.group(1), priv_ip.group(2))
                        if new_net_settings:
                            logging.warning("Confict of local subnets, trying to resolve...")
                            modem.set_local_nic(new_net_settings['ip'], new_net_settings['mask'],
                                                new_net_settings['dhcp'], new_net_settings['dhcp'])
                            raise NewSettinngs
                        """
                        ...
                    else:
                        logging.warning("Cannot get WAN IP adress from the LTE module")

            else:  # Not registered to operator
                if is_registered is True:
                    logging.info("Deregistered from network operator. {} Cycles to restart".format(reg_timeout))
                is_registered = False
                reg_timeout -= 1
                logging.debug("Dereg counter: {}".format(reg_timeout))
                if reg_timeout == 0:
                    reg_timeout = current_config['regtimeout']
                    raise TimeoutException("Cannot register to network")
                modem.modem.query("AT+CPIN?\r\n")  # Check da SIM

            logging.debug(f"RSSI: {rssi} sigqual: {sigqual} net-type: {nettype}")

            if not current_config['ping']['active'] or not operator:  # Manage ping
                modem.update_led(is_registered, rssi, WAN_STATE_MONITOR_OFF)
            elif ping_timestep:
                ping_timestep -= 1
                logging.debug(f"Next ping in {ping_timestep} cycles")
                modem.update_led(is_registered, rssi, conn_quard.get_wan_status())

            else:
                ping_timestep = current_config['ping']['period']
                ping_result = conn_quard.ping()
                if not conn_quard.get_rtable_valid():
                    ping_timestep = 0  # If rtable is invalid, try ping again a bit earlier

                if current_config['lte_autokeep_on'] is True:
                    conn_quard.keep_iface_on()
                modem.update_led(is_registered, rssi, conn_quard.get_wan_status())
                stats.report_metrics(rssi, sigqual, conn_quard.get_wan_status(), nettype, ping_result)

            if stats.is_time_fullfilled():  # Generate report to the log
                logging.info(stats.get_summmary(do_reset=True))

            stats.report_metrics(rssi, sigqual, conn_quard.get_wan_status(), nettype, ping_result=None)
            time.sleep(10 - ((time.time() - start_timestamp) % 10))
            logging.debug("STEP")

        except ModemCreateInstanceException as e:
            logging.error("Cannot create modem instance: " + str(e))
            do_action(None, ACTION_SUSPEND)
            reinit = do_action(None, ACTION_HARDRESTART)
            continue

        except ModemConnectionError as e:
            logging.error("Cannot connect to WAN via LTE, restarting...")
            reinit = do_action(modem, ACTION_HARDRESTART)
            continue

        except (ModemErrorException, TimeoutException) as e:
            action, errcode = modem.resolve_error(e)
            reinit = do_action(modem, action, errcode)
            logging.debug("Continuing")
            continue

        except NewSettings as e:
            logging.info("Settings of the modem has been changed, restarting...")
            reinit = do_action(modem, ACTION_SOFTRESTART)
            continue

        except KeyboardInterrupt:
            modem.leds.all_off()
            logging.shutdown()

except LEDErrorException as ER:
    traceback.print_exc()
    logging.critical("LED subsystem error:" + str(ER))

except Exception as ER:
    logging.critical("Unhandled exception: " + str(ER))

finally:
    del stats
    os.unlink(PIDFILE_PATH)
    if modem is not None:
        modem.do_cleanup()
